<?php
/**
  * Munin Graph JS
  * @author Dominique Fournier <dominique@fournier38.fr>
  * Licence BSD
  */

/** This class allow to manipulate de Munin files to extract the content.
  */
class Munin
{
  // PROPERTIES
  // {{{
  /** Datafile absolute path
    */
  private $DATAPATH = "/var/lib/munin";

  /** Configuration file
    */
  private $CONFFILE = "/etc/munin/munin.conf";

  /** The Datafile content in array format
    */
  private $datafileContent = null;

  /** The RRDCache daemon address
    */
  private $rrdcacheAddress = null;

  /** The limit file is cached
    */
  private $limitCache = null;

  /** The exceptions set in configuration file
    * Format :
    * array (site1/backupfournier.fournier38.fr/cpu/guest => array (
    *          warning =>
    *          critical =>
    *          unknown =>
    *       )
    */
  private $exceptions = null;
  // }}}

  // SETTERS/GETTERS
  /** Set / Get the data path
    * @param string|null $val The path of data path
    * @return string|$this
    */
  public function datapath ($val = null)
  // {{{
  {
    if ($val === null)
      return $this->DATAPATH;
    if (! file_exists ($val))
      throw new \Exception ("No datafile '$val' exists", 404);
    if (! is_readable ($val))
      throw new \Exception ("Datafile '$val' is not readable", 404);
    $this->DATAPATH = $val;
    return $this;
  }
  // }}}

  /** Set / Get the rrdcacheAddress
    * @param string|null $val The path of data path
    * @return string|$this
    */
  public function rrdcacheAddress ($val = null)
  // {{{
  {
    if ($val === null)
      return $this->rrdcacheAddress;
    $this->rrdcacheAddress = $val;
    return $this;
  }
  // }}}


  // METHODS
  /** Read the Datafile and put all in memory cache
    */
  public function readDataFile ()
  // {{{
  {
    $datafile = $this->DATAPATH."/datafile";
    if (! file_exists ($datafile) || ! is_readable ($datafile))
      throw new \Exception ("No datafile '$datafile' exists or ".
      "it is not readable", 500);

    // Analyzer of Datafile
    $data = array ();
    // Get the groups, hosts and modules
    foreach (file ($datafile) as $line)
    {
      $line  = trim ($line);
      @list ($group, $continue) = explode (";", $line);
      if ($continue === null)
        continue;
      @list ($host, $continue) = explode (":", $continue, 2);
      if ($continue === null)
        continue;
      @list ($module, $continue) = explode (".", $continue, 2);
      @list ($part, $val) = explode (" ", $continue, 2);
      @list ($internal, $part) = explode (".", $part, 2);
      @list ($field, $param) = explode (".", $part, 2);
      if ($field === null || $field === "")
        $data["$group/$host/$module"][$internal] = $val;
      elseif ($param === null || $param === "")
        $data["$group/$host/$module"][$internal][$field] = $val;
      else
        $data["$group/$host/$module"][$internal][$field][$param] = $val;
    }
    ksort ($data, SORT_NATURAL);
    // Overload the warning and critical from the /etc/munin/munin.conf file as
    // the datafile contain the default values.
    $confFile = file ($this->CONFFILE);
    if ($confFile === false)
      throw new \Exception ("Can not read the '$this->CONFFILE' file");
    $group = "";
    $host = "";
    foreach ($confFile as $line)
    {
      $line = trim ($line);
      if (! isset ($line{0}) || $line{0} === "#")
        continue;
      if ($line{0} === "[")
      {
        $line = substr ($line, 1, -1);
        list ($group, $host) = explode (";", $line);
      }
      else
      {
        @list ($object, $value) = explode (" ", $line);
        @list ($module, $internal, $priority) = explode (".", $object);
        if ($priority !== "unknown" &&
            $priority !== "warning" &&
            $priority !== "critical")
          continue;
        if (key_exists ("$group/$host/$module", $data) &&
            key_exists ($internal, $data["$group/$host/$module"]) &&
            key_exists ($priority, $data["$group/$host/$module"][$internal]))
        {
          $data["$group/$host/$module"][$internal][$priority] = $value;
        }
        $this->exceptions["$group/$host/$module/$internal"][$priority] =
          $value;
      }
    }
    $this->datafileContent = $data;
    return $data;
  }
  // }}}

  /** Return the tree group/host/module/internal defined in Datafile, in array
    * format
    * @return array[] = group/host/module/internal
    */
  public function groupsHostsModulesInternal ()
  // {{{
  {
    if ($this->datafileContent === null)
      $this->readDataFile ();
    $res = array ();
    foreach ($this->datafileContent as $key => $internals)
    {
      foreach ($internals as $internal => $internalParams)
      {
        if (substr ($internal, 0, 6) === "graph_")
          continue;
        $res[] = "$key/$internal";
      }
    }
    return $res;
  }
  // }}}

  /** Return the tree group/host/module defined in Datafile, in array format
    * @return array[] = group/host/module
    */
  public function groupsHostsModules ()
  // {{{
  {
    if ($this->datafileContent === null)
      $this->readDataFile ();
    return array_keys ($this->datafileContent);
  }
  // }}}

  /** Return the tree group/host  defined in Datafile, in array format
    * @return array[] = group/host
    */
  public function groupsHosts ()
  // {{{
  {
    $res = array ();
    foreach ($this->groupsHostsModules () as $name)
    {
      list ($group, $host, $module) = explode ("/", $name);
      $res[] = "$group/$host";
    }
    return array_unique ($res);
  }
  // }}}

  /** Return the last update for group/host
    * @return array[group/host] = timestamp
    */
  public function lastHostUpdate ()
  // {{{
  {
    $storageFiles = $this->DATAPATH."/state-*.storable";
    $groupsHosts = $this->groupsHosts ();
    $res = array ();
    foreach ($groupsHosts as $groupHost)
    {
      $groupHostTmp = str_replace ("/", "-", $groupHost);
      $file = $this->DATAPATH."/state-$groupHostTmp.storable";
      if (! file_exists ($file))
        continue;
      $res[$groupHost] = filemtime ($file);
    }
    return $res;
  }
  // }}}

  ////    VALUES MANAGEMENT    /////
  /** Read the RRD file values
    * @param string $nameInternal The RRD file to get
    * @param integer $start Timestamp to start
    * @param integer $end Timestamp to end
    * @param integer|null $resolution Resolution in seconds
    * @param string|null $cf AVERAGE, MIN, MAX, LAST
    * @return array
    */
  public function getRRD ($nameInternal, $start, $end, $resolution = 300,
    $cf = "AVERAGE")
  // {{{
  {
    if (substr ($nameInternal, -4) === ".rrd")
      $nameInternal = substr ($nameInternal, 0, -4);
    if (! in_array ($cf, ["AVERAGE", "MIN", "MAX", "LAST"]))
      throw new \Exception ("Munin getRRD CF : not allowed : ".
      var_export ($cf, true));
    @list ($group, $host, $module, $internal) = explode ("/", $nameInternal, 4);
    $nameInternal = $this->DATAPATH."/$group/$host-$module-$internal";
    $search = "$nameInternal-[cdg].rrd";
    $files = glob ($search);
    if (empty ($files))
      throw new \Exception ("Can not find '$search' file");
    if (count ($files) > 1)
      throw new \Exception ("More than one RRD file for $search");
    $nameInternal = reset ($files);
    $cmd = "rrdtool fetch '$nameInternal' '$cf' -s '$start' -e '$end' ".
           "-r '$resolution'";
    if ($this->rrdcacheAddress !== null && $this->rrdcacheAddress !== "")
      $cmd .= " -d '$this->rrdcacheAddress'";
    $cmd .= " 2>&1";
    //echo "// $cmd\n";
    exec ($cmd, $output, $rc);
    if ($rc !== 0)
      throw new \Exception ("Can not get data from RRD file '$nameInternal' : ".
      implode ("\n", $output));
    $res = array ();
    foreach ($output as $line)
    {
      @list ($timestamp, $value) = explode (":", $line);
      if ($value === "" || $value === null)
        continue;
      if (trim ($value) !== "-nan")
        $res[$timestamp] = (float) $value;
      else
        $res[$timestamp] = "";
    }
    return $res;
  }
  // }}}

  /** Get the values from RRD file and convert them if needed
    * @param string $name The name of  the module with slashes
    * @param string $internal The internal name to get
    * @param string $timeperiod The time period to get "day", "week", "month",
    *   "year"
    * @param string|null $cf AVERAGE, MIN, MAX, LAST
    * @return array
    */
  public function getValues ($name, $internal, $timeperiod, $cf = "AVERAGE")
  // {{{
  {
    if ($this->datafileContent === null)
      $this->readDataFile ();
    if (! key_exists ($name, $this->datafileContent))
      throw new \Exception ("Name '$name' not found in Datafile");
    $negatives = array ();
    switch ($timeperiod)
    {
      case "day":
        // 34 hours
        $startTimestamp = 34;
        $resolution = 300;
        break;
      case "week":
        // 9 days
        $startTimestamp = 24 * 9;
        $resolution = 1800;
        break;
      case "month":
        // 5 weeks : 35 days
        $startTimestamp = 24 * 35;
        $resolution = 7200;
        break;
      case "year":
        // 14 monthes
        $startTimestamp = 24 * 420;
        $resolution = 86400;
        break;
      default:
        http_response_code (403);
        throw new \Exception ("InvalidTimePeriod provided to graph : ".
        var_export ($timeperiod, true));
    }
    $start = time () - $startTimestamp * 3600;
    $end = time ();
    foreach ($this->datafileContent[$name] as $key => $vals)
    {
      if (! is_array ($vals))
        continue;
      if (key_exists ("negative", $vals))
        $negatives[] = $vals["negative"];
    }
    $values = $this->getRRD ("$name/$internal", $start, $end, $resolution, $cf);
    // Convert the timestamp to date and the negative values
    foreach ($values as $key => $val)
    {
      unset ($values[$key]);
      if ($val !== 0.0 && $val !== "")
      {
        if (in_array ($internal, $negatives))
          $val = -1*$val;
        if (key_exists ("cdef", $this->datafileContent[$name][$internal]))
        {
          @list ($internalNotUsed, $integer, $operator) = explode (",",
            $this->datafileContent[$name][$internal]["cdef"]);
          if ($operator === "*")
            $val *= $integer;
          elseif ($operator === "/")
            $val /= $integer;
          elseif ($operator === "+")
            $val += $integer;
          elseif ($operator === "-")
            $val -= $integer;
          else
            throw new \Exception ("CDEF Operator unknown :'$operator'");
        }
      }
      $values[date ("Y-m-d H:i:s", $key)] = $val;
    }
    //array_pop ($values);
    return $values;
  }
  // }}}


  ////    LIMIT MANAGEMENT    ////
  /** Read the limit file, extract it in array, and cache it
    * @return array
    */
  public function getLimits ()
  // {{{
  {
    if (! file_exists ("$this->DATAPATH/limits"))
      throw new \Exception ("File $this->DATAPATH/limits doesn't exists");
    if (! is_readable ("$this->DATAPATH/limits"))
      throw new \Exception ("File $this->DATAPATH/limits is not readable");
    $res = array ();
    foreach (file ("$this->DATAPATH/limits") as $line)
    {
      $line = trim ($line);
      if ($line === "")
        continue;
      if (strpos ($line, ";state ") !== false)
      {
        $line = str_replace (";state ", " ", $line);
        @list ($nameInternal, $state) = explode (" ", $line);
        $nameInternal = str_replace (";", "/", $nameInternal);
        $res[$nameInternal]["state"] = $state;
      }
      if (strpos ($line, ";unknown ") !== false)
      {
        $line = str_replace (";unknown ", " ", $line);
        @list ($nameInternal, $message) = explode (" ", $line, 2);
        $nameInternal = str_replace (";", "/", $nameInternal);
        $res[$nameInternal]["message"] = $message;
      }
      if (strpos ($line, ";warning ") !== false)
      {
        $line = str_replace (";warning ", " ", $line);
        @list ($nameInternal, $message) = explode (" ", $line, 2);
        $nameInternal = str_replace (";", "/", $nameInternal);
        $res[$nameInternal]["message"] = $message;
      }
      if (strpos ($line, ";critical ") !== false)
      {
        $line = str_replace (";critical ", " ", $line);
        @list ($nameInternal, $message) = explode (" ", $line, 2);
        $nameInternal = str_replace (";", "/", $nameInternal);
        $res[$nameInternal]["message"] = $message;
      }
    }
    $this->limitCache = $res;
    return $res;
  }
  // }}}

  /** Get the internals with the state in Unknown
    * @return array ($nameInternal => $message);
    */
  public function getUnknown ()
  // {{{
  {
    if ($this->limitCache === null)
      $this->getLimits ();
    if ($this->datafileContent === null)
      $this->readDataFile ();
    // Here the exceptions are set, if exists
    $res = array ();
    foreach ($this->limitCache as $key => $vals)
    {
      if ($vals["state"] !== "unknown")
        continue;
      if ($this->exceptions !== null &&
          key_exists ($key, $this->exceptions) &&
          key_exists ("unknown", $this->exceptions[$key]) &&
          $this->exceptions[$key]["unknown"] === "1")
        continue;
      $res[$key] = $vals["message"];
    }
    return $res;
  }
  // }}}

  /** Get the internals with the state in Warning
    * @return array ($nameInternal => $message);
    */
  public function getWarning ()
  // {{{
  {
    if ($this->limitCache === null)
      $this->getLimits ();
    $res = array ();
    foreach ($this->limitCache as $key => $vals)
    {
      if ($vals["state"] !== "warning")
        continue;
      $res[$key] = $vals["message"];
    }
    return $res;
  }
  // }}}

  /** Get the internals with the state in Critical
    * @return array ($nameInternal => $message);
    */
  public function getCritical ()
  // {{{
  {
    if ($this->limitCache === null)
      $this->getLimits ();
    $res = array ();
    foreach ($this->limitCache as $key => $vals)
    {
      if ($vals["state"] !== "critical")
        continue;
      $res[$key] = $vals["message"];
    }
    return $res;
  }
  // }}}
}
